
const arg_qselall = 'querySelectorAll';
const querySelectorAll = document[arg_qselall].bind(document);
const [ el_html ] = querySelectorAll('html');

const _define = (key, name, value) => {
	Object.defineProperty(key, name, { value });
};

const isObject = (value) => typeof value === 'object' && null !== value;
const iterateObject = (value) => Object.entries(value);

// Creating NodeList is a big deal
// https://stackoverflow.com/questions/13351966/create-node-list-from-a-single-node-in-javascript
// DO NOT create NodeList from DocumentFragment — it's remove nodes from document and append them to the DocumentFragment
const createNodeList = (elements) => {
	if (elements instanceof Set === false) {
		elements = new Set(elements);
	}
	elements = [ ...elements ];

	const node_list = document.createDocumentFragment().childNodes;

	const params = {};
	for (const [ index, element ] of elements.entries()) {
		params[index] = {
			value: element,
			enumerable: true,
		};
	}
	params.length = { value: elements.length };
	params.item = {
		value (i) {
			return this[+i || 0]; // eslint-disable-line no-implicit-coercion
		},
		enumerable: true,
	};

	// return an object pretending to be a NodeList
	return Object.create(node_list, params);
};
const parseHTML = (html) => {
	// const tmp = document.implementation.createHTMLDocument();
	// tmp.body.innerHTML = html;
	const tmp = document.createElement('div');
	tmp.innerHTML = html;

	for (const el of tmp.$find('script')) {
		const el_new = document.createElement('script');
		{
			const attrs = el.$attrAll();
			for (const [ name, value ] of iterateObject(attrs)) {
				el_new.$attr(
					name,
					value,
				);
			}
		}
		el_new.innerHTML = el.innerHTML;
		el.$replaceWith(el_new);
	}
	return tmp.childNodes.$freeze();
};

const addToSet = (set, ...args) => {
	for (const one of args) {
		set.add(one);
	}
};

const parseOnArgs = (events_list, ...args) => {
	const events = new Set(events_list.split(' '));
	let selector = null;
	let callback_given = null;
	let options = {};
	if (typeof args[0] === 'function') {
		[ callback_given, options ] = args;
	}
	else if (typeof args[0] === 'string' && typeof args[1] === 'function') {
		[ selector, callback_given, options ] = args;
	}

	if (null === callback_given) {
		throw new Error('Invalid arguments.');
	}

	return { events, selector, callback_given, options };
};

const pel = {};
const plist = {};
const pwin = {};

// Extend prototypes
{
	const matchesSelector = (el, selector) => (undefined === selector) ? true : (el.matches || el.matchesSelector || el.msMatchesSelector || el.mozMatchesSelector || el.webkitMatchesSelector || el.oMatchesSelector).call(el, selector);

	const argsToNodes = (args, text_mode = 0) => {
		const nodes = [];
		if (Array.isArray(args[0])) {
			args = args[0];
		}
		for (const arg of args) {
			if (arg instanceof Element || arg instanceof Node) {
				nodes.push(arg);
			}
			else if (arg instanceof NodeList) {
				nodes.push(...arg);
			}
			else if (typeof arg === 'string') {
				if (arg[0] === '<') {
					nodes.push(...parseHTML(arg));
				}
				else {
					if (1 === text_mode) {
						nodes.push(document.createTextNode(arg));
					}
					else {
						nodes.push(
							...querySelectorAll(arg),
						);
					}
				}
			}
		}
		return nodes;
	};

	const dashedToCamel = (key) => {
		return key.replace(/-([a-z])/, (_, letter) => letter.toUpperCase());
	};

	// Matched elements matching / change
	// is, add, filter, not
	// TODO has
	{
		plist.is = function (selector) {
			return this[0] ? matchesSelector(this[0], selector) : false;
		};

		plist.add = function (...args) {
			return createNodeList([ ...this, ...argsToNodes(args) ]);
		};

		const doFilter = (elements, filter, op = true) => {
			const nodes = [];
			for (const el of elements) {
				if (matchesSelector(el, filter) === op) {
					nodes.push(el);
				}
			}
			return createNodeList(nodes);
		};
		plist.filter = function (filter) {
			return doFilter(this, filter);
		};
		plist.not = function (filter) {
			return doFilter(this, filter, false);
		};
	}

	// DOM traversal
	// find, childs, parent, parents, parentsUntil, next, nextAll, prev, prevAll, siblings, contains
	// TODO nextUntil, prevUntil, contents
	{
		pel.find = function (selector) {
			return this[arg_qselall](selector);
		};
		plist.find = function (selector) {
			const nodes = [];
			for (const el of this) {
				nodes.push(...el[arg_qselall](selector));
			}
			return createNodeList(nodes);
		};

		plist.childs = plist.children = function (selector) {
			const need_filter = undefined !== selector;
			const nodes = [];
			for (const el of this) {
				if (need_filter) {
					for (const el_child of el.children) {
						if (matchesSelector(el_child, selector)) {
							nodes.push(el_child);
						}
					}
				}
				else {
					nodes.push(...el.children);
				}
			}
			return createNodeList(nodes);
		};

		const getParents = (el, filter, opts = {}) => {
			const { until = null, recursive = true } = opts;
			const nodes = new Set();

			const el_parent = el.parentNode;
			if (!!el_parent && el_parent !== document) { // eslint-disable-line no-implicit-coercion
				if (null === until || !matchesSelector(el_parent, until)) {
					if (matchesSelector(el_parent, filter)) {
						nodes.add(el_parent);
					}
					if (recursive) {
						addToSet(
							nodes,
							...getParents(
								el_parent,
								filter,
								opts,
							),
						);
					}
				}
			}

			return nodes;
		};
		plist.parent = function (filter) {
			const nodes = new Set();
			for (const el of this) {
				addToSet(
					nodes,
					...getParents(
						el,
						filter,
						{ recursive: false },
					),
				);
			}
			return createNodeList(nodes);
		};
		plist.parents = function (filter) {
			const nodes = new Set();
			for (const el of this) {
				addToSet(
					nodes,
					...getParents(el, filter),
				);
			}
			return createNodeList(nodes);
		};
		plist.parentsUntil = function (selector, filter) {
			const nodes = new Set();
			for (const el of this) {
				addToSet(
					nodes,
					...getParents(
						el,
						filter,
						{
							until: selector,
						},
					),
				);
			}
			return createNodeList(nodes);
		};

		const getNeighborElements = (el, selector, params = {}) => {
			const nodes = [];
			while (1) { // eslint-disable-line no-constant-condition
				el = el[(true === params.prev ? 'previous' : 'next') + 'ElementSibling'];
				if (null === el || undefined === el) {
					break;
				}
				else if (undefined === selector || matchesSelector(el, selector)) {
					nodes.push(el);
					if (true === params.one) {
						break;
					}
				}
			}
			return nodes;
		};
		plist.next = function (selector) {
			const nodes = [];
			for (const el of this) {
				const els_next = getNeighborElements(
					el,
					selector,
					{
						one: true,
					},
				);
				if (els_next.length > 0) {
					nodes.push(
						els_next[0],
					);
				}
			}
			return createNodeList(nodes);
		};
		plist.nextAll = function (selector) {
			const nodes = [];
			for (const el of this) {
				const els_next = getNeighborElements(el, selector);
				nodes.push(
					...els_next,
				);
			}
			return createNodeList(nodes);
		};
		plist.prev = function (selector) {
			const nodes = [];
			for (const el of this) {
				const els_next = getNeighborElements(
					el,
					selector,
					{
						prev: true,
						one: true,
					},
				);

				if (els_next.length > 0) {
					nodes.push(
						els_next[0],
					);
				}
			}
			return createNodeList(nodes);
		};
		plist.prevAll = function (selector) {
			const nodes = [];
			for (const el of this) {
				const els_next = getNeighborElements(
					el,
					selector,
					{ prev: true },
				);
				nodes.push(
					...els_next,
				);
			}
			return createNodeList(nodes);
		};

		plist.siblings = function (selector) {
			const nodes = new Set();
			for (const el of this) {
				for (const el_sibling of el.parentNode.children) {
					if (
						el_sibling !== el &&
						(
							undefined === selector ||
							matchesSelector(el_sibling, selector)
						)
					) {
						nodes.add(el_sibling);
					}
				}
			}
			return createNodeList(nodes);
		};

		plist.index = function () {
			return this[0] ? this[0].$prevAll().length : null;
		};

		plist.contains = function (selector) {
			return this[0] ? this[0].querySelector(selector) !== null : false;
		};
	}

	// DOM operations
	// append, appendTo, prepend, prependTo, insertBefore, before, insertAfter, after, remove, detach, clone, replaceWith
	{
		{
			const ops = [
				(target, el) => {
					target.append(el);
				},
				(target, el) => {
					target.insertBefore(el, target.firstChild);
				},
				(target, el) => {
					target.parentNode.insertBefore(el, target);
				},
				(target, el) => {
					const parent = target.parentNode;
					const next = target.nextSibling;
					if (next) {
						next.before(el);
					}
					else {
						ops[0](parent, el);
					}
				},
			];
			const addTo = (op_type, targets, elements, order = 0) => {
				const fn = ops[op_type];

				elements = argsToNodes(elements, +!order); // eslint-disable-line no-implicit-coercion
				const targets_copy = createNodeList(targets);

				const args = [];
				for (const el of targets_copy) {
					args[order] = el;
					for (const node of elements) {
						args[+!order] = node; // eslint-disable-line no-implicit-coercion
						fn(...args);
					}
				}

				return targets_copy;
			};
			plist.append = function (...args) {
				return addTo(0, this, args);
			};
			plist.appendTo = function (...args) {
				return addTo(0, this, args, 1);
			};
			plist.prepend = function (...args) {
				return addTo(1, this, args);
			};
			plist.prependTo = function (...args) {
				return addTo(1, this, args, 1);
			};
			plist.before = function (...args) {
				return addTo(2, this, args);
			};
			plist.insertBefore = function (...args) {
				return addTo(2, this, args, 1);
			};
			plist.after = function (...args) {
				return addTo(3, this, args);
			};
			plist.insertAfter = function (...args) {
				return addTo(3, this, args, 1);
			};

			plist.replaceWith = function (content) {
				if (typeof content === 'string') {
					for (const el of this) {
						el.outerHTML = content;
					}
				}
				else if (content instanceof Element || content instanceof NodeList) {
					addTo(3, this, [ content ]);
					return this.$detach();
				}
				return this;
			};
		}

		// pel.remove = function(){
		// 	this.parentNode.removeChild(this);
		// };
		// plist.remove = function(){
		// 	for(const el of this){
		// 		el.parentNode.removeChild(el);
		// 	}
		// };
		const detach = (node) => {
			node.remove();
			return node;
		};
		pel.detach = pel.remove = function () {
			return detach(this);
		};
		plist.detach = plist.remove = function () {
			const nodes = [];
			for (const el of this) {
				nodes.push(
					detach(el),
				);
			}
			return createNodeList(nodes);
		};

		pel.clone = function () {
			return this.cloneNode(true);
		};
		plist.clone = function () {
			const nodes = [];
			for (const el of this) {
				nodes.push(
					el.cloneNode(true),
				);
			}
			return createNodeList(nodes);
		};
	}

	// Element size & position
	// width, innerWidth, outerWidth, height, innerHeight, outerHeight, offset, position
	{
		const getWidthWithPadding = (el) => el ? el.clientWidth : undefined;
		plist.width = function () {
			const [ el ] = this;
			if (el) {
				const styles = getComputedStyle(el);
				return getWidthWithPadding(el) - parseInt(styles.paddingLeft) - parseInt(styles.paddingRight);
			}
		};
		plist.innerWidth = function () {
			return getWidthWithPadding(this[0]);
		};
		plist.outerWidth = function (with_margin) {
			const [ el ] = this;
			if (el) {
				let result = el.offsetWidth;
				if (true === with_margin) {
					const styles = getComputedStyle(el);
					result += parseInt(styles.marginLeft) + parseInt(styles.marginRight);
				}
				return result;
			}
		};

		const getHeightWithPadding = (el) => el ? el.clientHeight : undefined;
		plist.height = function () {
			const [ el ] = this;
			if (el) {
				const styles = getComputedStyle(el);
				return getHeightWithPadding(el) - parseInt(styles.paddingTop) - parseInt(styles.paddingBottom);
			}
		};
		plist.innerHeight = function () {
			return getHeightWithPadding(this[0]);
		};
		plist.outerHeight = function (with_margin) {
			const [ el ] = this;
			if (el) {
				let result = el.offsetHeight;
				if (true === with_margin) {
					const styles = getComputedStyle(el);
					result += parseInt(styles.marginTop) + parseInt(styles.marginBottom);
				}
				return result;
			}
		};

		plist.offset = function () {
			if (this[0]) {
				const rect = this[0].getBoundingClientRect();
				return {
					top: rect.top + document.body.scrollTop,
					left: rect.left + document.body.scrollLeft,
				};
			}
		};
		plist.position = function () {
			const [ el ] = this;
			if (el) {
				return {
					left: el.offsetLeft,
					top: el.offsetTop,
				};
			}
		};
	}

	// Element content:
	// html, text, empty
	{
		plist.html = function (...args) {
			if (args.length === 0) {
				return this.length > 0 ? this[0].innerHTML : undefined;
			}
			else {
				for (const el of this) {
					el.innerHTML = args[0];
				}
				return this;
			}
		};
		plist.text = function (...args) {
			if (args.length === 0) {
				return this.length > 0 ? this[0].textContent : undefined;
			}
			else {
				for (const el of this) {
					el.textContent = args[0];
				}
				return this;
			}
		};
		plist.empty = function () {
			for (const el of this) {
				el.innerHTML = '';
			}
			return this;
		};
	}

	// CSS
	// classAdd, classRemove, classAdded, css, hide, unhide
	{
		plist.classAdd = plist.addClass = function (...class_list) {
			for (const el of this) {
				el.classList.add(...class_list);
			}
			return this;
		};
		plist.classRemove = plist.removeClass = function (...class_list) {
			for (const el of this) {
				el.classList.remove(...class_list);
			}
			return this;
		};
		plist.classAdded = plist.hasClass = function (class_name) {
			return this[0] ? this[0].classList.contains(class_name) : false;
		};

		plist.css = function (...args_styles) {
			let styles = null;
			if (1 === args_styles.length) {
				if (null !== args_styles[0] && isObject(args_styles[0])) {
					styles = args_styles[0];
				}
				else if (typeof args_styles[0] === 'string') {
					return getComputedStyle(this[0])[args_styles[0]];
				}
			}
			else if (2 === args_styles.length && typeof args_styles[0] === 'string') {
				styles = {
					[ args_styles[0] ]: args_styles[1],
				};
			}

			if (null === styles) {
				throw new Error('Invalid arguments.');
			}

			for (let [ key, value ] of iterateObject(styles)) {
				if (typeof value === 'number') {
					value += 'px';
				}
				const key_css = dashedToCamel(key);

				for (const el of this) {
					el.style[key_css] = value;
				}
			}

			return this;
		};

		plist.hide = function () {
			for (const el of this) {
				el.style.display = 'none';
			}
			return this;
		};
		plist.unhide = function () {
			for (const el of this) {
				el.style.display = '';
			}
			return this;
		};
	}

	// Attributes:
	// attr, attrRemove, prop, data, value
	{
		const getAttr = (el, name) => el.hasAttribute(name) ? el.getAttribute(name) : null;
		plist.attr = function (...args) {
			let attrs = {};
			if (isObject(args[0])) {
				attrs = args[0];
			}
			else if (typeof args[0] === 'string') {
				const [ name, value ] = args;
				if (1 === args.length) {
					return this.length > 0 ? getAttr(this[0], name) : null;
				}
				else {
					attrs[name] = value;
				}
			}
			else {
				throw new TypeError('Invalid type given.');
			}

			for (const el of this) {
				for (const [ key, value ] of iterateObject(attrs)) {
					el.setAttribute(
						key,
						value,
					);
				}
			}

			return this;
		};
		plist.attrAll = function () {
			const el = this[0];
			const attrs = {};
			for (const name of el.getAttributeNames()) {
				const value = getAttr(el, name);
				if (value) {
					attrs[name] = value;
				}
			}
			return attrs;
		};
		plist.attrRemove = plist.removeAttr = function (name) {
			for (const el of this) {
				el.removeAttribute(name);
			}
			return this;
		};

		plist.prop = function (...args) {
			const [ name, value ] = args;
			if (1 === args.length) {
				return this.length > 0 ? this[0][name] : undefined;
			}
			else {
				for (const el of this) {
					el[name] = value;
				}
				return this;
			}
		};

		// data
		{
			const data_key = '$_datastorage';
			plist.data = function (...args) {
				let datas = {};
				if (isObject(args[0])) {
					datas = args[0];
				}
				else if (typeof args[0] === 'string') {
					const [ key, value ] = args;
					if (1 === args.length) {
						return this.length > 0 && this[0][data_key] ? this[0][data_key].get(dashedToCamel(key)) : undefined;
					}
					else {
						datas[key] = value;
					}
				}
				else {
					throw new TypeError('Invalid type given');
				}
				for (const el of this) {
					if (!el[data_key]) {
						el[data_key] = new Map();
					}

					for (const [ key, value ] of iterateObject(datas)) {
						el[data_key].set(
							dashedToCamel(key),
							value,
						);
					}
				}
				return this;
			};
			plist.dataRemove = plist.removeData = function (name) {
				for (const el of this) {
					if (el[data_key]) {
						el[data_key].delete(name);
					}
				}
				return this;
			};
		}

		plist.value = plist.val = function (...args) {
			if (args.length === 0) {
				return this.length > 0 ? this[0].value : undefined;
			}
			else {
				for (const el of this) {
					el.value = args[0];
				}
				return this;
			}
		};
	}

	// Events
	// on, once, off, trigger
	{
		const callbacks_storage = new Map();
		const op_bind = (op, els, args) => {
			const {
				events,
				selector,
				callback_given,
				options = {},
			} = parseOnArgs(...args);

			let callback = null;
			if ('off' === op) {
				callback = callbacks_storage.get(callback_given);
				callbacks_storage.delete(callback_given);
			}
			else {
				callback = function (ev) {
					if (null === selector) {
						callback_given.call(this, ev);
					}
					else {
						const current_targets = new Set();

						if (matchesSelector(ev.target, selector)) {
							current_targets.add(ev.target);
						}

						for (const el_current_target of ev.target.$parents(selector)) {
							current_targets.add(el_current_target);
						}

						for (const el_current_target of current_targets) {
							callback_given.call(
								el_current_target,
								ev,
								{
									currentTarget: el_current_target,
								},
							);
						}
					}
					if ('once' === op) {
						op_bind('off', els, args);
					}
				};
			}

			callbacks_storage.set(callback_given, callback);

			const listener_pref = ('off' === op ? 'remove' : 'add');
			op_do(els, listener_pref, events, callback, options);

			return els;
		};
		const op_do = (els, listener_pref, events, callback, options) => {
			for (const el of els) {
				for (const ev of events) {
					el[listener_pref + 'EventListener'](ev, callback, options);
				}
			}
		};
		pwin.on = function (...args) {
			return op_bind('on', this, args);
		};
		pwin.once = function (...args) {
			return op_bind('once', this, args);
		};
		pwin.off = function (...args) {
			return op_bind('off', this, args);
		};

		const native_events = new Set();
		for (const name of Object.keys(HTMLElement.prototype)) {
			if (name.startsWith('on')) {
				native_events.add(
					name.slice(2),
				);
			}
		}
		const createEvent = (
			name,
			detail = {},
			options = {},
		) => {
			const is_native = native_events.has(name.toLowerCase());
			// console.log('is_native', is_native);

			const {
				bubbles = true,
				cancelable = true,
			} = options;

			if (is_native) {
				return new Event(
					name,
					{
						bubbles,
						cancelable,
					},
				);
			}
			else {
				return new CustomEvent(
					name,
					{
						bubbles,
						cancelable,
						detail,
					},
				);
			}
		};
		pwin.trigger = function (name, detail) {
			const ev = createEvent(name, detail);
			for (const el of this) {
				el.dispatchEvent(ev);
			}

			return this;
		};
	}

	// Service
	// feeeze
	{
		// returns static (non-live) copy of NodeList
		// "In some cases, the NodeList is live, which means that changes in the DOM automatically update the collection. For example, Node.childNodes is live"
		// https://developer.mozilla.org/en-US/docs/Web/API/NodeList
		plist.freeze = function () {
			return createNodeList(this);
		};
	}
}

// general functions
const $man = (arg) => {
	if (typeof arg === 'string') {
		arg = arg.trim();
		if ('<' === arg[0]) {
			return parseHTML(arg);
		}
		else {
			return querySelectorAll(arg);
		}
	}
	else if (typeof arg === 'function') {
		if (document.attachEvent ? document.readyState === 'complete' : document.readyState !== 'loading') {
			arg();
		}
		else {
			document.addEventListener('DOMContentLoaded', arg);
		}
	}
	else if (arg instanceof Element) {
		return createNodeList([ arg ]);
	}
	else if (arg instanceof NodeList) {
		return arg;
	}
};

const def_man = (name, get) => {
	Object.defineProperty($man, name, { get });
};

def_man(
	'parseHTML',
	() => parseHTML,
);

def_man(
	'head',
	() => document.head,
);
def_man(
	'body',
	() => document.body,
);

const addMethod = (name, fn, fn_single, fn_window) => {
	name = '$' + name;

	fn = fn || fn_window;
	if (!fn_single) {
		fn_single = function (...args) {
			return fn.apply(
				createNodeList([ this ]),
				args,
			);
		};
	}

	_define(
		Element.prototype,
		name,
		fn_single,
	);

	_define(
		$man,
		name,
		fn_single.bind(el_html),
	);

	_define(
		NodeList.prototype,
		name,
		fn,
	);

	if (fn_window) {
		_define(
			window,
			name,
			fn_single.bind(window),
		);
		_define(
			document,
			name,
			fn_single.bind(document),
		);
	}
};

for (const [ name, fn_plist ] of iterateObject(plist)) {
	addMethod(
		name,
		fn_plist,
		pel[name],
	);
}
for (const [ name, fn_pwin ] of iterateObject(pwin)) {
	addMethod(
		name,
		plist[name],
		pel[name],
		fn_pwin,
	);
}

def_man(
	'extend',
	() => (...args) => {
		addMethod(...args);
	},
);

// mouseenter, mouseleave
{
	const isSelfOrParentMatch = (el_given, selector) => {
		if (!el_given || !el_given.$is) {
			// do nothing
		}
		else if (el_given.$is(selector)) {
			return el_given;
		}
		else {
			return el_given.$parents(selector)[0];
		}
	};
	const createFn = (els, name, ...args) => {
		const { selector, callback_given, options } = parseOnArgs('', ...args);

		const callback = (ev) => {
			// console.log(ev.target, ev.relatedTarget);
			const el_target = isSelfOrParentMatch(ev.target, selector);
			const el_related = isSelfOrParentMatch(ev.relatedTarget, selector);
			// console.log(el_target, el_related);
			if (el_target && el_target !== el_related) {
				callback_given(ev);
			}
		};

		for (const el of els) {
			el.addEventListener(
				'mouse' + name,
				callback,
				options,
			);
		}

		return els;
	};
	addMethod(
		'onmouseenter',
		function (...args) {
			return createFn(
				this,
				'over',
				...args,
			);
		},
	);
	addMethod(
		'onmouseleave',
		function (...args) {
			return createFn(
				this,
				'out',
				...args,
			);
		},
	);
}

export default $man;
